import os
import argparse
import yaml

import pypeliner
import pypeliner.workflow
import pypeliner.managed as mgd

import destruct.benchmark.align.bwa.workflow
import destruct.benchmark.destruct_test
import destruct.benchmark.create_breakpoint_simulation
import destruct.benchmark.generate_bam


if __name__ == '__main__':

    argparser = argparse.ArgumentParser()
    pypeliner.app.add_arguments(argparser)

    argparser.add_argument('sim_config',
                           help='Simulation configuration filename')

    argparser.add_argument('ref_data_dir',
                           help='Reference genomes directory')

    argparser.add_argument('results_dir',
                           help='Output directory')

    argparser.add_argument('tool_defs',
                           help='Tool Definition Filename')

    args = vars(argparser.parse_args())

    try:
        os.makedirs(args['results_dir'])
    except OSError:
        pass

    args['tmpdir'] = os.path.join(args['results_dir'], 'tmp')

    pyp = pypeliner.app.Pypeline(config=args)

    yaml_text = open(args['tool_defs']).read().format(ref_data_dir=args['ref_data_dir'])
    tool_defs = yaml.load(yaml_text)

    sim_config = yaml.load(open(args['sim_config']))

    ctx = {'mem':4}

    workflow = pypeliner.workflow.Workflow(default_ctx=ctx)

    workflow.setobj(mgd.TempOutputObj('simulation.params'), sim_config['simulation'])
    workflow.setobj(mgd.TempOutputObj('chromosomes'), sim_config['reference']['chromosomes'])
    workflow.setobj(mgd.TempOutputObj('include_nonchromosomal'), sim_config['reference']['include_nonchromosomal'])

    workflow.subworkflow(
        name='generate_bam_workflow',
        func=destruct.benchmark.generate_bam.generate_bam,
        args=(
            mgd.TempInputObj('simulation.params'),
            mgd.TempInputObj('chromosomes'),
            mgd.TempInputObj('include_nonchromosomal'),
            mgd.OutputFile(os.path.join(args['results_dir'], 'simulated.bam')),
            mgd.OutputFile(os.path.join(args['results_dir'], 'genome.fasta')),
            mgd.OutputFile(os.path.join(args['results_dir'], 'simulated.tsv')),
            os.path.join(args['results_dir'], 'raw'),
        ),
    )

    workflow.setobj(
        obj=mgd.TempOutputObj('tool_defs', 'tool_name'),
        value=tool_defs,
    )

    workflow.subworkflow(
        name='run_tool',
        axes=('tool_name',),
        func=destruct.benchmark.destruct_test.create_tool_workflow,
        args=(
            mgd.TempInputObj('tool_defs', 'tool_name'),
            {'simulated': mgd.InputFile(os.path.join(args['results_dir'], 'simulated.bam')),},
            mgd.OutputFile(os.path.join(args['results_dir'], 'results_{tool_name}.tsv'), 'tool_name'),
            mgd.TempSpace('tool_raw_data', 'tool_name', cleanup='none'),
        ),
    )

    workflow.transform(
        name='plot',
        axes=('tool_name',),
        func=destruct.benchmark.destruct_test.create_roc_plot,
        args=(
            mgd.TempInputObj('simulation.params'),
            mgd.TempInputObj('tool_defs', 'tool_name'),
            mgd.InputFile(os.path.join(args['results_dir'], 'genome.fasta')),
            mgd.InputFile(os.path.join(args['results_dir'], 'simulated.tsv')),
            mgd.InputFile(os.path.join(args['results_dir'], 'results_{tool_name}.tsv'), 'tool_name'),
            mgd.OutputFile(os.path.join(args['results_dir'], 'annotated_{tool_name}.tsv'), 'tool_name'),
            mgd.OutputFile(os.path.join(args['results_dir'], 'identified_{tool_name}.tsv'), 'tool_name'),
            mgd.OutputFile(os.path.join(args['results_dir'], 'plots_{tool_name}.pdf'), 'tool_name'),
        ),
    )

    pyp.run(workflow)



import pandas as pd


def destruct_postprocess(breakpoint_table_filename, breakpoint_library_table_filename, output_filename, control_id=None):

    breakpoint_table = pd.read_csv(breakpoint_table_filename, sep='\t')
    breakpoint_library_table = pd.read_csv(breakpoint_library_table_filename, sep='\t')

    breakpoint_counts = (
        breakpoint_library_table
        .set_index(['prediction_id', 'library'])[['num_reads']]
        .unstack(fill_value=0)
    )

    breakpoint_counts.columns = [a[1] + '_count' for a in breakpoint_counts.columns]

    breakpoint_table = breakpoint_table.merge(breakpoint_counts,
                                              left_on='prediction_id',
                                              right_index=True)
    
    # Filter based on evidence in control dataset
    if control_id is not None:
         breakpoint_table = breakpoint_table[breakpoint_table['{0}_count'.format(control_id)] == 0]          

    breakpoint_table.to_csv(output_filename, sep='\t', index=False)




import numpy as np

import destruct.balanced


def annotate_results(brks):
    # Classify balanced rearrangements as inversions, reciprocal translocations, complex
    balanced_rearrangements = destruct.balanced.detect_balanced_rearrangements(brks)

    brks.set_index('prediction_id', inplace=True)
    brks['rearrangement_type'] = ''

    for rearrangement in balanced_rearrangements:
        pred_types = brks.loc[rearrangement.prediction_ids, 'type'].values
        
        if len(rearrangement.prediction_ids) == 2 and np.all(pred_types == 'inversion'):
            brks.loc[rearrangement.prediction_ids, 'rearrangement_type'] == 'inversion'
            
        elif len(rearrangement.prediction_ids) == 2 and np.all(pred_types == 'translocation'):
            brks.loc[rearrangement.prediction_ids, 'rearrangement_type'] == 'reciprocal'

        else:
            brks.loc[rearrangement.prediction_ids, 'rearrangement_type'] == 'complex'

    brks.reset_index(inplace=True)

    # Classify non inversions with small distance as foldbacks
    brks['dist'] = np.absolute(brks['position_1'] - brks['position_2'])
    is_foldback = (brks['dist'] < 1000) & (brks['type'] == 'inversion') & (brks['rearrangement_type'] != 'inversion')
    brks.loc[is_foldback, 'rearrangement_type'] = 'foldback'

    # All non-complex deletions as deletions
    is_deletion = (brks['type'] == 'deletion') & (brks['rearrangement_type'] != 'complex')
    brks.loc[is_deletion, 'rearrangement_type'] = 'deletion'

    # All non-complex duplications as duplications
    is_duplication = (brks['type'] == 'duplication') & (brks['rearrangement_type'] != 'complex')
    brks.loc[is_duplication, 'rearrangement_type'] = 'duplication'

    # All non-complex non reciprocal translocations as translocations
    is_translocation = (brks['type'] == 'translocation') & (brks['rearrangement_type'] != 'complex') & (brks['rearrangement_type'] != 'reciprocal')
    brks.loc[is_translocation, 'rearrangement_type'] = 'translocation'

    # All non-complex non balanced inversions as unbalanced
    is_unbalanced = (brks['type'] == 'inversion') & (brks['rearrangement_type'] != 'complex') & (brks['rearrangement_type'] != 'inversion') & (brks['rearrangement_type'] != 'foldback')
    brks.loc[is_unbalanced, 'rearrangement_type'] = 'unbalanced'

    return brks




from __future__ import absolute_import
import itertools
import io
import tarfile
import scipy
import scipy.stats
import numpy as np


class gaussian_kde_set_covariance(scipy.stats.gaussian_kde):
    def __init__(self, dataset, covariance):
        self.covariance = covariance
        scipy.stats.gaussian_kde.__init__(self, dataset)
    def _compute_covariance(self):
        self.inv_cov = 1.0 / self.covariance
        self._norm_factor = np.sqrt(2*np.pi*self.covariance) * self.n


def filled_density(ax, data, c, a, xmin, xmax, cov):
    density = gaussian_kde_set_covariance(data, cov)
    xs = [xmin] + list(np.linspace(xmin, xmax, 2000)) + [xmax]
    ys = density(xs)
    ys[0] = 0.0
    ys[-1] = 0.0
    ax.plot(xs, ys, color=c, alpha=a)
    ax.fill(xs, ys, color=c, alpha=a)


def filled_density_weighted(ax, data, weights, c, a, xmim, xmax, cov):
    samples = np.array(list(itertools.chain(*[itertools.repeat(idx, cnt) for idx, cnt in enumerate(np.random.multinomial(10000, weights / weights.sum()))])))
    filled_density(ax, data[samples], c, a, xmim, xmax, cov)


def savefig_tar(tar, fig, filename):
    plot_buffer = io.BytesIO()
    fig.savefig(plot_buffer, format='pdf')
    info = tarfile.TarInfo(name=filename)
    info.size = plot_buffer.tell()
    plot_buffer.seek(0)
    tar.addfile(tarinfo=info, fileobj=plot_buffer)





import pypeliner.commandline


def concatenate_bcf(in_files, out_file):
    """ Fast concatenation of BCF file using `bcftools`.
    """
    
    cmd = ['bcftools', 'concat', '-a', '-O', 'b', '-o', out_file]
    cmd += [in_files[x] for x in sorted(in_files.keys())]
    
    pypeliner.commandline.execute(*cmd)
                



import csv
import os

import pypeliner

import destruct.defaultconfig


def wget_gunzip(url, filename):
    temp_filename = filename + '.tmp'
    pypeliner.commandline.execute('wget', url, '-c', '-O', temp_filename + '.gz')
    pypeliner.commandline.execute('gunzip', temp_filename + '.gz')
    os.rename(temp_filename, filename)

def wget(url, filename):
    temp_filename = filename + '.tmp'
    pypeliner.commandline.execute('wget', url, '-c', '-O', temp_filename)
    os.rename(temp_filename, filename)

class AutoSentinal(object):
    def __init__(self, sentinal_prefix):
        self.sentinal_prefix = sentinal_prefix
    def run(self, func):
        sentinal_filename = self.sentinal_prefix + func.__name__
        if os.path.exists(sentinal_filename):
            return
        func()
        with open(sentinal_filename, 'w'):
            pass

def create_ref_data(config, ref_data_dir):
    config = destruct.defaultconfig.get_config(ref_data_dir, config)

    try:
        os.makedirs(ref_data_dir)
    except OSError:
        pass

    auto_sentinal = AutoSentinal(ref_data_dir + '/sentinal.')

    temp_directory = os.path.join(ref_data_dir, 'tmp')

    try:
        os.makedirs(temp_directory)
    except OSError:
        pass

    def wget_genome_fasta():
        with open(config['genome_fasta'], 'w') as genome_file:
            for assembly in config['ensembl_assemblies']:
                assembly_url = config['ensembl_assembly_url'].format(assembly)
                assembly_fasta = os.path.join(temp_directory, 'dna.assembly.{0}.fa'.format(assembly))
                if not os.path.exists(assembly_fasta):
                    wget_gunzip(assembly_url, assembly_fasta)
                with open(assembly_fasta, 'r') as assembly_file:
                    for line in assembly_file:
                        if line[0] == '>':
                            line = line.split()[0] + '\n'
                        genome_file.write(line)
    auto_sentinal.run(wget_genome_fasta)

    def wget_gtf():
        wget_gunzip(config['ensembl_gtf_url'], config['gtf_filename'])
    auto_sentinal.run(wget_gtf)

    def wget_dgv():
        wget(config['dgv_url'], config['dgv_filename'])
    auto_sentinal.run(wget_dgv)

    def wget_repeats():
        repeat_filename = os.path.join(temp_directory, 'repeats.txt')
        wget_gunzip(config['rmsk_url'], repeat_filename)
        with open(config['chromosome_map'], 'r') as chr_map_file:
            chr_map = dict((a.split() for a in chr_map_file))
        with open(repeat_filename, 'r') as repeat_file, open(config['repeat_regions'], 'w') as repeat_regions_file, open(config['satellite_regions'], 'w') as satellite_regions_file:
            for row in csv.reader(repeat_file, delimiter='\t'):
                if row[5] not in chr_map:
                    continue
                chr = chr_map[row[5]]
                start = str(int(row[6]) + 1)
                end = row[7]
                type = row[11]
                repeat_regions_file.write('\t'.join([chr, start, end]) + '\n')
                if type == 'Satellite':
                    satellite_regions_file.write('\t'.join([chr, start, end]) + '\n')
    auto_sentinal.run(wget_repeats)

    def bowtie_build():
        pypeliner.commandline.execute('bowtie-build', config['genome_fasta'], config['genome_fasta'])
    auto_sentinal.run(bowtie_build)

    def samtools_faidx():
        pypeliner.commandline.execute('samtools', 'faidx', config['genome_fasta'])
    auto_sentinal.run(samtools_faidx)





import argparse
import ConfigParser
import random
import string
import os
import subprocess
import sys

import utils.seq
import utils.misc


class SequenceRegion(object):
    def __init__(self, chromosome, strand, start, end):
        self.chromosome = chromosome
        self.strand = strand
        self.start = start
        self.end = end
    def __repr__(self):
        return 'SequenceRegion({0},{1},{2},{3})'.format(repr(self.chromosome), repr(self.strand), repr(self.start), repr(self.end))
    def __eq__(self, other):
        return self.chromosome == other.chromosome and self.strand == other.strand and self.start == other.start and self.end == other.end
    @property
    def length(self):
        return self.end - self.start + 1
    def subregion(self, local_start, local_end):
        return SequenceRegion(self.chromosome, self.strand, self.start + local_start - 1, self.start + local_end - 1)
    def invert(self):
        strand = ('+', '-')[self.strand == '+']
        assert strand != self.strand
        return SequenceRegion(self.chromosome, strand, self.start, self.end)

class TumourChromosome(object):
    def __init__(self, regions, genome):
        self.regions = regions
        self.length = sum([a.end - a.start + 1 for a in regions])
        self.genome = genome
    def local_start_ends(self):
        """
        generate a list of regions, in addition to the start and end
        position of the region in the tumour chromosome
        """
        offset = 0
        for region in self.regions:
            yield region, offset + 1, offset + region.length
            offset += region.length
    def reference_position(self, position):
        """ given a position in the tumour chromosome, identify the position
        in the reference genome
        """
        assert position <= self.length
        offset = 0
        for region in self.regions:
            if position <= offset + region.length:
                region_position = position - offset
                if region.strand == '-':
                    region_position = region.length - region_position + 1
                return region.chromosome, region.strand, region.start + region_position - 1
            offset += region.length
    def generate_spaced_intervals(self, num, size_func, spacing=1000):
        """ generate a set of non overlapping intervals in the tumour
        chromosome such that there is at least spacing nucleotides between
        intervals and at least spacing nucleotides between interval boundaries
        and existing distinct sequence regions that make up the tumour
        chromosome
        """
        intervals = list()
        while len(intervals) < num:
            size = size_func()
            start = random.randint(1 + spacing, self.length - size - spacing - 1)
            end = start + size - 1
            overlapping = False
            for info in intervals:
                if start <= info[1] + spacing and end >= info[0] - spacing:
                    overlapping = True
                    break
            for region, local_start, local_end in self.local_start_ends():
                if abs(start - local_start) < spacing or abs(start - local_end) < spacing or abs(end - local_start) < spacing or abs(end - local_end) < spacing:
                    overlapping = True
                    break
            if overlapping:
                continue
            unmappable = False
            for chromosome_position in (start, end):
                chromosome, strand, position = self.reference_position(chromosome_position)
                if self.genome[chromosome].count('N', position - 100, position + 100) > 0:
                    unmappable = True
            if unmappable:
                continue
            intervals.append((start, end))
        return sorted(intervals)
    def split_regions(self, intervals):
        """ split the current tumour chromosome regions at the boundaries
        of the given set of intervals
        """
        for interval in intervals:
            new_regions = list()
            for region, local_start, local_end in self.local_start_ends():
                start_ends = list([1, region.length])
                if interval[0] > local_start and interval[0] < local_end:
                    start_ends.append(interval[0] - local_start)
                    start_ends.append(interval[0] - local_start + 1)
                if interval[1] > local_start and interval[1] < local_end:
                    start_ends.append(interval[1] - local_start + 1)
                    start_ends.append(interval[1] - local_start + 2)
                start_ends = sorted(start_ends)
                for local_start, local_end in zip(start_ends[::2], start_ends[1::2]):
                    new_regions.append(region.subregion(local_start, local_end))
            self.regions = sorted(new_regions, key=lambda a: a.start)
    def merge_regions(self):
        """ merge adjacent tumour chromosome regions if those regions are adjacent
        and consecutive in the reference genome
        """
        new_regions = list()
        idx1 = 0
        while idx1 < len(self.regions):
            idx2 = idx1 + 1
            while idx2 < len(self.regions):
                if self.regions[idx1].chromosome != self.regions[idx2].chromosome:
                    break
                if self.regions[idx1].strand != self.regions[idx2].strand:
                    break
                if self.regions[idx2-1].end != self.regions[idx2].start - 1:
                    break
                idx2 += 1
            idx2 -= 1
            new_regions.append(SequenceRegion(self.regions[idx1].chromosome, self.regions[idx1].strand, self.regions[idx1].start, self.regions[idx2].end))
            idx1 = idx2 + 1
        self.regions = new_regions
        assert self.length == sum([a.end - a.start + 1 for a in new_regions])
    def simulate_deletions(self, num, size_func, spacing=1000):
        """ simulate deletions of size given by size_func spaced by spacing
        """
        deleted = self.generate_spaced_intervals(num, size_func, spacing)
        self.split_regions(deleted)
        new_regions = list()
        for region, local_start, local_end in self.local_start_ends():
            if len(deleted) == 0 or local_end < deleted[0][0]:
                new_regions.append(region)
            elif local_end <= deleted[0][1]:
                pass
            elif local_end > deleted[0][1]:
                deleted = deleted[1:]
                new_regions.append(region)
        self.regions = new_regions
        self.length = sum([a.end - a.start + 1 for a in new_regions])
        self.merge_regions()
    def simulate_duplications(self, num, size_func, spacing=1000):
        """ simulate duplications of size given by size_func spaced by spacing
        """
        duplicated = self.generate_spaced_intervals(num, size_func, spacing)
        self.split_regions(duplicated)
        new_regions = list()
        dup_regions = list()
        for region, local_start, local_end in self.local_start_ends():
            if len(duplicated) == 0 or local_end < duplicated[0][0]:
                new_regions.append(region)
            elif local_end <= duplicated[0][1]:
                dup_regions.append(region)
            elif local_end > duplicated[0][1]:
                new_regions.extend(dup_regions)
                new_regions.extend(dup_regions)
                dup_regions = list()
                duplicated = duplicated[1:]
                new_regions.append(region)
        self.regions = new_regions
        self.length = sum([a.end - a.start + 1 for a in new_regions])
        self.merge_regions()
    def simulate_inversions(self, num, size_func, spacing=1000):
        """ simulate inversions of size given by size_func spaced by spacing
        """
        inverted = self.generate_spaced_intervals(num, size_func, spacing)
        self.split_regions(inverted)
        new_regions = list()
        inv_regions = list()
        for region, local_start, local_end in self.local_start_ends():
            if len(inverted) == 0 or local_end < inverted[0][0]:
                new_regions.append(region)
            elif local_end <= inverted[0][1]:
                inv_regions.append(region)
            elif local_end > inverted[0][1]:
                new_regions.extend([a.invert() for a in reversed(inv_regions)])
                inv_regions = list()
                inverted = inverted[1:]
                new_regions.append(region)
        self.regions = new_regions
        self.length = sum([a.end - a.start + 1 for a in new_regions])
        self.merge_regions()
    def create_sequence(self, genome):
        sequence = ''
        for region in self.regions:
            region_sequence = genome[region.chromosome][region.start-1:region.end]
            if region.strand == '-':
                region_sequence = utils.misc.reverse_complement(region_sequence)
            sequence += region_sequence
        return sequence
    @property
    def breakpoints(self):
        for id in range(len(self.regions)-1):
            region1 = self.regions[id]
            region2 = self.regions[id+1]
            if region1.strand == '+':
                pos1 = region1.end
                str1 = '+'
            else:
                pos1 = region1.start
                str1 = '-'
            if region2.strand == '+':
                pos2 = region2.start
                str2 = '-'
            else:
                pos2 = region2.end
                str2 = '+'
            yield id, region1.chromosome, str1, pos1, region2.chromosome, str2, pos2

def random_chimera(fasta, chromosome, num_chimera, fragment_length):
    id = 0
    chimera_length = 0
    while id < num_chimera:
        chimera = ''
        for side in (0, 1):
            start = random.randint(0, len(chromosome) - fragment_length - 1)
            end = start + fragment_length
            sequence = chromosome[start:end]
            if random.choice([True, False]):
                sequence = utils.misc.reverse_complement(sequence)
            chimera += sequence
        if chimera.count('N') > 0:
            continue
        fasta.write('>r{0}\n{1}\n'.format(id, chimera))
        id += 1
        chimera_length += len(chimera)
    return chimera_length

def write_fasta(fasta, id, sequence):
    nt_per_line = 80
    fasta.write('>' + str(id) + '\n')
    for start in range(0, len(sequence), nt_per_line):
        fasta.write(sequence[start:start+nt_per_line] + '\n')

def simulate(cfg, sim_info, read_count, sequences_fasta, reads1, reads2, random_reads=True):
    print ('Simulating {0} reads from {1}'.format(read_count, sequences_fasta))
    temp_prefix = reads1 + '.dwgsim.temp.prefix'
    random_reads_flag = ''
    if not random_reads:
        random_reads_flag = '-y 0'
    dwgsim_command = cfg.dwgsim_bin + ' -H {7} -z {0} -d {1} -s {2} -N {3} -1 {4} -2 {4} {5} {6}'.format(sim_info['dwgsim_seed'], sim_info['fragment_mean'], sim_info['fragment_stddev'], read_count, sim_info['read_length'], sequences_fasta, temp_prefix, random_reads_flag)
    print (dwgsim_command)
    dwgsim_retcode = subprocess.call(dwgsim_command.split())
    assert(dwgsim_retcode == 0)
    os.rename(temp_prefix + '.bwa.read1.fastq', reads1)
    os.rename(temp_prefix + '.bwa.read2.fastq', reads2)
    try:
        os.remove(temp_prefix + '.bfast.fastq')
    except OSError:
        pass
    try:
        os.remove(temp_prefix + '.mutations.txt')
    except OSError:
        pass
    try:
        os.remove(temp_prefix + '.mutations.vcf')
    except OSError:
        pass

def create(cfg, sim_info, chromosomes_info, breakpoints_info, reads_fastq_1, reads_fastq_2, temps_prefix):
    random.seed(int(sim_info['rearrangements_seed']))
    genome = dict(utils.seq.read_sequences(open(cfg.genome_fasta, 'r')))
    simulation_chromosome = sim_info['simulation_chromosome']
    # Create simulated chromosomes by applying random operations
    tumour_proportions = [float(a) for a in sim_info['chromosome_proportions'].split()]
    simulated_chromosomes = list()
    for chromosome_id, proportion in enumerate(tumour_proportions):
        simulated_chromosome = TumourChromosome([SequenceRegion(simulation_chromosome, '+', 1, len(genome[simulation_chromosome]))], genome)
        simulated_chromosome.simulate_deletions(int(sim_info['deletion_count']), lambda: random.randint(*[int(a) for a in sim_info['deletion_range'].split()]))
        simulated_chromosome.simulate_duplications(int(sim_info['duplication_count']), lambda: random.randint(*[int(a) for a in sim_info['duplication_range'].split()]))
        simulated_chromosome.simulate_inversions(int(sim_info['inversion_count']), lambda: random.randint(*[int(a) for a in sim_info['inversion_range'].split()]))
        simulated_chromosomes.append(simulated_chromosome)
    # Write chromosome regions info
    with open(chromosomes_info, 'w') as info:
        for chromosome_id, simulated_chromosome in enumerate(simulated_chromosomes):
            for region in simulated_chromosome.regions:
                info.write('\t'.join([str(a) for a in (chromosome_id, region.chromosome, region.strand, region.start, region.end)]) + '\n')
    # Write breakpoint positions info
    with open(breakpoints_info, 'w') as info:
        for chromosome_id, simulated_chromosome in enumerate(simulated_chromosomes):
            for break_info in simulated_chromosome.breakpoints:
                info.write('\t'.join([str(a) for a in (chromosome_id, tumour_proportions[chromosome_id]) + break_info]) + '\n')
    create_sequences(cfg, sim_info, genome, simulated_chromosomes, tumour_proportions, reads_fastq_1, reads_fastq_2, temps_prefix)

def create_sequences(cfg, sim_info, genome, chromosomes, proportions, fastq1_filename, fastq2_filename, temps_prefix):
    # Generate fastas from simulated chromosomes
    chromosome_fastas = dict()
    for chromosome_id in range(len(chromosomes)):
        chromosome_fastas[chromosome_id] = temps_prefix + str(chromosome_id) + '.fa'
        with open(chromosome_fastas[chromosome_id], 'w') as fasta:
            write_fasta(fasta, 't'+str(chromosome_id), chromosomes[chromosome_id].create_sequence(genome))
    # Simulate reads from each tumour chromosome
    fastqs1 = list()
    fastqs2 = list()
    for chromosome_id in range(len(chromosomes)):
        scaled_coverage = proportions[chromosome_id] * float(sim_info['coverage'])
        scaled_read_count = int(float(scaled_coverage) * chromosomes[chromosome_id].length / float(sim_info['fragment_mean']))
        temp_fastq1_filename = temps_prefix + str(chromosome_id) + '.1.fastq'
        temp_fastq2_filename = temps_prefix + str(chromosome_id) + '.2.fastq'
        simulate(cfg, sim_info, scaled_read_count, chromosome_fastas[chromosome_id], temp_fastq1_filename, temp_fastq2_filename)
        fastqs1.append(temp_fastq1_filename)
        fastqs2.append(temp_fastq2_filename)
    # Simulate random chimeric fragments
    false_chimera_fasta = temps_prefix + 'false_chimera.fa'
    with open(false_chimera_fasta, 'w') as fasta:
        chimera_length = random_chimera(fasta, genome[sim_info['simulation_chromosome']], int(sim_info['num_false_chimera']), int(sim_info['fragment_mean']))
    false_chimera_fastq1 = temps_prefix + 'f.1.fastq'
    false_chimera_fastq2 = temps_prefix + 'f.2.fastq'
    fastqs1.append(false_chimera_fastq1)
    fastqs2.append(false_chimera_fastq2)
    false_chimera_read_count = int(float(sim_info['false_chimera_coverage']) * chimera_length / float(sim_info['fragment_mean']))
    simulate(cfg, sim_info, false_chimera_read_count, false_chimera_fasta, false_chimera_fastq1, false_chimera_fastq2, False)
    # Concatenate fastqs
    with open(fastq1_filename, 'w') as fastq1:
        subprocess.check_call(['cat'] + fastqs1, stdout=fastq1)
    with open(fastq2_filename, 'w') as fastq2:
        subprocess.check_call(['cat'] + fastqs2, stdout=fastq2)
    for fastq in fastqs1 + fastqs2:
        os.remove(fastq)




import pypeliner
import pypeliner.managed as mgd

import destruct.workflow
import destruct.benchmark.wrappers.utils
import destruct.benchmark.wrappers.lumpysv.tasks


def create_lumpysv_wrapper_workflow(bam_filenames, output_filename, raw_data_dir, control_id=None):
    bams = list()
    for lib_id, bam_filename in bam_filenames.items():
        bams += [destruct.benchmark.wrappers.utils.symlink(bam_filename, link_name='{0}.bam'.format(lib_id), link_directory=raw_data_dir)]
        destruct.benchmark.wrappers.utils.symlink(bam_filename+'.bai', link_name='{0}.bam.bai'.format(lib_id), link_directory=raw_data_dir)

    workflow = pypeliner.workflow.Workflow(default_ctx={'mem': 4})

    workflow.setobj(
        obj=mgd.OutputChunks('sample_id'),
        value=list(bam_filenames.keys()),
    )

    workflow.commandline(
        name='extract_discordants',
        axes=('sample_id',),
        args=(
            'samtools', 'view', '-b', '-F', '1294',
            mgd.InputFile('bam', 'sample_id', fnames=bam_filenames),
            '>',
            mgd.TempOutputFile('discordants.unsorted.bam', 'sample_id'),
        ),
    )

    workflow.commandline(
        name='extract_splitters',
        axes=('sample_id',),
        args=(
            'samtools', 'view', '-h',
            mgd.InputFile('bam', 'sample_id', fnames=bam_filenames),
            '|',
            'lumpy_extractSplitReads_BwaMem', '-i', 'stdin',
            '|',
            'samtools', 'view', '-Sb', '-',
            '>',
            mgd.TempOutputFile('splitters.unsorted.bam', 'sample_id'),
        ),
    )

    workflow.commandline(
        name='sort_discordants',
        axes=('sample_id',),
        args=(
            'samtools', 'sort',
            mgd.TempInputFile('discordants.unsorted.bam', 'sample_id'),
            '-o',
            mgd.TempOutputFile('discordants.bam', 'sample_id'),
        ),
    )

    workflow.commandline(
        name='sort_splitters',
        axes=('sample_id',),
        args=(
            'samtools', 'sort',
            mgd.TempInputFile('splitters.unsorted.bam', 'sample_id'),
            '-o',
            mgd.TempOutputFile('splitters.bam', 'sample_id'),
        ),
    )

    workflow.transform(
        name='run_lumpyexpress',
        func=destruct.benchmark.wrappers.lumpysv.tasks.run_lumpyexpress,
        args=(
            mgd.InputFile('bam', 'sample_id', fnames=bam_filenames),
            mgd.TempInputFile('splitters.bam', 'sample_id'),
            mgd.TempInputFile('discordants.bam', 'sample_id'),
            mgd.TempOutputFile('results.vcf'),
        ),
    )

    workflow.transform(
        name='vcf_to_bcf',
        func=destruct.benchmark.wrappers.lumpysv.tasks.vcf_to_bcf,
        args=(
            mgd.TempInputFile('results.vcf'),
            mgd.TempOutputFile('results.bcf'),
        ),
    )

    workflow.transform(
        name='convert_bcf',
        func=destruct.benchmark.wrappers.lumpysv.tasks.convert_bcf,
        ctx={'mem': 4, 'num_retry': 3, 'mem_retry_increment': 2},
        args=(
            mgd.TempInputFile('results.bcf'),
            mgd.OutputFile(output_filename),
        ),
        kwargs={
            'control_id': control_id,
        },
    )

    return workflow


def setup_lumpysv(test_config, **kwargs):
    pass



from setuptools import setup, find_packages
import versioneer

setup(
    name='ngs-destruct',
    packages=find_packages(),
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Destruct is a tool for joint prediction of rearrangement breakpoints from single or multiple tumour samples',
    author='Andrew McPherson',
    author_email='andrew.mcpherson@gmail.com',
    url='http://bitbucket.org/dranew/destruct',
    download_url='https://bitbucket.org/dranew/destruct/get/v{}.tar.gz'.format(versioneer.get_version()),
    keywords=['scientific', 'sequence analysis', 'cancer'],
    classifiers=[],
    entry_points={'console_scripts': ['destruct = destruct.ui.main:main']},
    package_data={'destruct': ['data/*_chr_map.tsv']},
    setup_requires=[
        'numpy',
        'setuptools',
    ],
    install_requires=[
        'numpy',
        'pandas',
        'pypeliner',
        'pygenes',
        'pyyaml',
        'matplotlib',
        'seaborn',
        'pysam',
        'blossomv',
    ],
    dependency_links=[
        'https://github.com/amcpherson/blossomv/archive/v2.04_r4.tar.gz#egg=blossomv',
    ],
)




import os
import argparse
import yaml

import pypeliner
import pypeliner.workflow
import pypeliner.managed as mgd

import destruct.benchmark.align.bwa.workflow
import destruct.benchmark.destruct_test
import destruct.benchmark.create_breakpoint_simulation


def generate_bam(
    simulation_params,
    chromosomes,
    include_nonchromosomal,
    simulated_bam_filename,
    genome_fasta_filename,
    simulated_table_filename,
    raw_data_dir,
):
    workflow = pypeliner.workflow.Workflow(default_ctx={'mem':4})

    workflow.setobj(mgd.TempOutputObj('simulation.params'), simulation_params)
    workflow.setobj(mgd.TempOutputObj('chromosomes'), chromosomes)
    workflow.setobj(mgd.TempOutputObj('include_nonchromosomal'), include_nonchromosomal)

    workflow.transform(
        name='create_genome',
        func=destruct.benchmark.destruct_test.create_genome,
        args=(
            mgd.TempInputObj('chromosomes'),
            mgd.TempInputObj('include_nonchromosomal'),
            mgd.OutputFile(genome_fasta_filename),
        ),
    )

    workflow.transform(
        name='create_sim',
        func=destruct.benchmark.create_breakpoint_simulation.create,
        args=(
            mgd.TempInputObj('simulation.params'),
            mgd.InputFile(genome_fasta_filename),
            mgd.OutputFile(os.path.join(raw_data_dir, 'simulated.fasta')),
            mgd.OutputFile(simulated_table_filename),
            mgd.TempOutputFile('concordant.1.fastq'),
            mgd.TempOutputFile('concordant.2.fastq'),
            mgd.TempOutputFile('discordant.1.fastq'),
            mgd.TempOutputFile('discordant.2.fastq'),
        ),
    )

    workflow.commandline(
        name='cat1',
        args=(
            'cat',
            mgd.TempInputFile('concordant.1.fastq'),
            mgd.TempInputFile('discordant.1.fastq'),
            '>', mgd.OutputFile(os.path.join(raw_data_dir, 'simulated.1.fastq')),
        ),
    )

    workflow.commandline(
        name='cat2',
        args=(
            'cat',
            mgd.TempInputFile('concordant.2.fastq'),
            mgd.TempInputFile('discordant.2.fastq'),
            '>', mgd.OutputFile(os.path.join(raw_data_dir, 'simulated.2.fastq')),
        ),
    )

    workflow.subworkflow(
        name='bwa_align',
        func=destruct.benchmark.align.bwa.workflow.bwa_align_workflow,
        args=(
            mgd.InputFile(genome_fasta_filename),
            mgd.InputFile(os.path.join(raw_data_dir, 'simulated.1.fastq')),
            mgd.InputFile(os.path.join(raw_data_dir, 'simulated.2.fastq')),
            mgd.TempOutputFile('simulated.unsorted.bam'),
        ),
    )

    workflow.transform(
        name='samtools_sort_index',
        func=destruct.benchmark.destruct_test.samtools_sort_index,
        args=(
            mgd.TempInputFile('simulated.unsorted.bam'),
            mgd.OutputFile(simulated_bam_filename),
        ),
    )

    return workflow


if __name__ == '__main__':

    argparser = argparse.ArgumentParser()
    pypeliner.app.add_arguments(argparser)

    argparser.add_argument('sim_config',
                           help='Simulation configuration filename')

    argparser.add_argument('ref_data_dir',
                           help='Reference genomes directory')

    argparser.add_argument('raw_data_dir',
                           help='Output directory')

    argparser.add_argument('bam_filename',
                           help='Output bam filename')

    argparser.add_argument('genome_fasta',
                           help='Output genome fasta filename')

    argparser.add_argument('simulated_table',
                           help='Output table of breakpoints in TSV format')

    args = vars(argparser.parse_args())

    try:
        os.makedirs(args['raw_data_dir'])
    except OSError:
        pass

    args['tmpdir'] = os.path.join(args['raw_data_dir'], 'tmp')

    pyp = pypeliner.app.Pypeline(config=args)

    sim_config = yaml.load(open(args['sim_config']))

    ctx = {'mem':4}

    workflow = generate_bam(
        sim_config['simulation'],
        sim_config['reference']['chromosomes'],
        sim_config['reference']['include_nonchromosomal'],
        args['bam_filename'],
        args['genome_fasta'],
        args['simulated_table'],
        args['raw_data_dir'],
    )

    pyp.run(workflow)



################################################
# Default configuration for destruct/demix
################################################

import pkg_resources

def get_config(ref_data_dir, user_config):

    ###
    # Gene annotations and reference genome retrieval
    ###

    # Version of ensembl for gene annotations
    ensembl_version                             = '93'

    # Associated genome version used by the ensembl version
    ensembl_genome_version                      = 'GRCh38'

    # Ensemble assemblies to include in the reference genome
    ensembl_assemblies                          = ['chromosome.1', 'chromosome.2', 'chromosome.3', 'chromosome.4', 'chromosome.5', 'chromosome.6', 'chromosome.7', 'chromosome.8', 'chromosome.9', 'chromosome.10', 'chromosome.11', 'chromosome.12', 'chromosome.13', 'chromosome.14', 'chromosome.15', 'chromosome.16', 'chromosome.17', 'chromosome.18', 'chromosome.19', 'chromosome.20', 'chromosome.21', 'chromosome.22', 'chromosome.X', 'chromosome.Y', 'chromosome.MT', 'nonchromosomal']

    # Base chromosomes, used for parallelization and by demix
    chromosomes                                 = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', 'X']

    # Ucsc genome version (must match ensembl version!)
    ucsc_genome_version                         = 'hg38'

    # Database of genomic variants genome version (must match ensembl version!)
    dgv_genome_version                          = 'GRCh38_hg38'

    # Database of genomic variants release
    dgv_version                                 = '2016-05-15'

    # Name of the mitochondrial chromosome in ensembl
    mitochondrial_chromosome                    = 'MT'

    ###
    # URLs of gene annotations and reference genome
    ###

    ensembl_assembly_url = 'ftp://ftp.ensembl.org/pub/release-'+ensembl_version+'/fasta/homo_sapiens/dna/Homo_sapiens.'+ensembl_genome_version+'.dna.{0}.fa.gz'

    ensembl_gtf_url = 'ftp://ftp.ensembl.org/pub/release-'+ensembl_version+'/gtf/homo_sapiens/Homo_sapiens.'+ensembl_genome_version+'.'+ensembl_version+'.gtf.gz'

    dgv_url = 'http://dgv.tcag.ca/dgv/docs/'+dgv_genome_version+'_variants_'+dgv_version+'.txt'

    if ucsc_genome_version == 'hg18':
        rmsk_url = 'ftp://hgdownload.cse.ucsc.edu/goldenPath/hg18/database/*_rmsk.txt.gz'
    else:
        rmsk_url = 'ftp://hgdownload.cse.ucsc.edu/goldenPath/'+ucsc_genome_version+'/database/rmsk.txt.gz'

    ###
    # Filenames of gene annotations and reference genome
    ###

    genome_fasta                                = ref_data_dir+'/Homo_sapiens.'+ensembl_genome_version+'.'+ensembl_version+'.dna.chromosomes.fa'
    genome_fai                                  = genome_fasta+'.fai'
    gtf_filename                                = ref_data_dir+'/Homo_sapiens.'+ensembl_genome_version+'.'+ensembl_version+'.gtf'
    dgv_filename                                = ref_data_dir+'/dgv.txt'
    repeat_regions                              = ref_data_dir+'/repeats.regions'
    satellite_regions                           = ref_data_dir+'/repeats.satellite.regions'

    # Mapping between ensembl and ucsc chromosome names, hg19 and hg18 provided for you
    chromosome_map                              = pkg_resources.resource_filename('destruct', 'data/'+ucsc_genome_version+'_chr_map.tsv')

    ###
    # Algorithm parameters
    ###

    # Maximum inferred fragment length of a read pair classified as concordant
    bam_max_fragment_length                     = 1000

    # Maximum soft clipped bases before a read is called discordant
    bam_max_soft_clipped                        = 8

    # Maximum fragment length used during realignemnt
    fragment_length_max                         = 800

    # Number of standard deviations to define min and max fragment length more precisely during realignment
    fragment_length_num_stddevs                 = 6.0

    # Minimum alignment probability for filtering discordant read alignments
    alignment_threshold                         = 0.1

    # Prior probability a read is chimeric
    chimeric_prior                              = 0.01

    # Minimum chimeric probability for filtering discordant read alignments
    chimeric_threshold                          = 0.1

    # Minimum read valid probability for filtering discordant read alignments
    readvalid_threshold                         = 0.01

    # Number of concordant reads sampled to calculate valid alignment score distribution
    num_read_samples                            = 100000

    # Minimum discordant read count for filtering clusters
    cluster_readcount_threshold                 = 2

    # Minimum template length aligned to either side of the breakpoint
    template_length_min_threshold               = 40

    # Minimum mate score for possibly concordant alignments
    mate_score_threshold                        = 60

    # Realignment parameters
    match_score                                 = 2
    mismatch_score                              = -3
    gap_score                                   = -4

    # Min alignment likelihood
    min_alignment_log_likelihood                = -5.0

    # Minimum score for a complex rearrangement cycle
    cycles_scoremax                             = 4

    # Minimum number of breakpoints to visit in search for complex rearrangement cycles
    cycles_visitmax                             = 100000

    # Lambda parameter for complex rearrangement discovery (see nFuse, McPherson et al. 2012)
    cycles_lambda                               = 2000

    ###
    # Simulation test parameters
    ###

    reads_per_job                               = 1000000

    ###
    # Parallelization parameters
    ###

    # Number of reads per parallel realignment job
    reads_per_split                             = 1000000

    # Number of clusters per parallel 
    clusters_per_split                          = 1000

    config = locals()
    del config['user_config']

    config.update(user_config)

    return config




import collections
import csv
import errno
import itertools
import os
import tarfile
import gzip
import numpy as np
import pandas as pd
import pypeliner
import pygenes

import destruct.utils.plots
import destruct.utils.seq
import destruct.predict_breaks


def prepare_seed_fastq(reads_1_fastq, reads_2_fastq, seed_length, seed_fastq):
    opener_1 = (open, gzip.open)[reads_1_fastq.endswith('.gz')]
    opener_2 = (open, gzip.open)[reads_2_fastq.endswith('.gz')]
    with opener_1(reads_1_fastq, 'rt') as reads_1, opener_2(reads_2_fastq, 'rt') as reads_2, open(seed_fastq, 'wt') as seed:
        fastq_lines = [[], []]
        for fastq_1_line, fastq_2_line in zip(reads_1, reads_2):
            fastq_lines[0].append(fastq_1_line.rstrip())
            fastq_lines[1].append(fastq_2_line.rstrip())
            if len(fastq_lines[0]) == 4:
                for read_end in (0, 1):
                    if len(fastq_lines[read_end][1]) > seed_length:
                        fastq_lines[read_end][1] = fastq_lines[read_end][1][0:seed_length]
                        fastq_lines[read_end][3] = fastq_lines[read_end][3][0:seed_length]
                    for line in fastq_lines[read_end]:
                        seed.write(line + '\n')
                fastq_lines = [[], []]


class ConcordantReadStats(object):
    def __init__(self, stats, fragment_length_num_stddevs):
        self.stats = stats
        self.fragment_length_num_stddevs = fragment_length_num_stddevs

    @property
    def fragment_length_mean(self):
        return float(self.stats['fragment_mean'])

    @property
    def fragment_length_stddev(self):
        return float(self.stats['fragment_stddev'])

    @property
    def fragment_length_min(self):
        return int(self.fragment_length_mean - self.fragment_length_num_stddevs * self.fragment_length_stddev)

    @property
    def fragment_length_max(self):
        return int(self.fragment_length_mean + self.fragment_length_num_stddevs * self.fragment_length_stddev)


def read_stats(stats_filename, fragment_length_num_stddevs):
    stats = pd.read_csv(stats_filename, sep='\t')
    flen_stats = stats.loc[stats['type'] == 'fragment_length'].drop('type', axis=1)
    flen_stats = flen_stats.astype(float)
    fragment_count = flen_stats['value'].sum()
    fragment_mean = (flen_stats['key'] * flen_stats['value']).sum() / fragment_count
    fragment_variance = ((flen_stats['key'] - fragment_mean) * (flen_stats['key'] - fragment_mean) * flen_stats['value']).sum() / (fragment_count - 1)
    fragment_stddev = fragment_variance**0.5
    return ConcordantReadStats({'fragment_mean': fragment_mean, 'fragment_stddev': fragment_stddev}, fragment_length_num_stddevs)


def write_stats_table(library_ids, lib_stats, stats_table_filename):
    with open(stats_table_filename, 'wt') as stats_table_file:
        for lib_name, library_id in library_ids.items():
            stats_table_file.write(str(library_id) + '\t')
            stats_table_file.write(str(lib_stats[lib_name].fragment_length_mean) + '\t')
            stats_table_file.write(str(lib_stats[lib_name].fragment_length_stddev) + '\n')


def split_file_byline(in_filename, lines_per_file, out_filename_callback):
    with open(in_filename, 'rt') as in_file:
        file_number = 0
        out_file = None
        out_file_lines = None
        try:
            for line in in_file:
                if out_file is None or out_file_lines == lines_per_file:
                    if out_file is not None:
                        out_file.close()
                    out_file = open(out_filename_callback(file_number), 'wt')
                    out_file_lines = 0
                    file_number += 1
                out_file.write(line)
                out_file_lines += 1
        finally:
            if out_file is not None:
                out_file.close()


def split_fastq(in_filename, num_reads_per_file, out_filename_callback):
    with gzip.open(in_filename, 'rt') as in_file:
        file_number = 0
        out_file = None
        out_file_read_count = None
        try:
            for name, seq, comment, qual in itertools.zip_longest(*[in_file]*4):
                if out_file is None or out_file_read_count == num_reads_per_file:
                    if out_file is not None:
                        out_file.close()
                    out_file = open(out_filename_callback(file_number), 'wt')
                    out_file_read_count = 0
                    file_number += 1
                out_file.write(name)
                out_file.write(seq)
                out_file.write(comment)
                out_file.write(qual)
                out_file_read_count += 1
        finally:
            if out_file is not None:
                out_file.close()


def merge_files_by_line(in_filenames, out_filename):
    with open(out_filename, 'wt') as out_file:
        for id, in_filename in sorted(in_filenames.items()):
            with open(in_filename, 'rt') as in_file:
                for line in in_file:
                    out_file.write(line)


def create_library_ids(library_names):
    return dict([(library_name, library_id) for library_id, library_name in enumerate(library_names)])


def merge_alignment_files(in_filenames, out_filename, library_idxs):
    with open(out_filename, 'wt') as out_file:
        for lib_id, in_filename in in_filenames.items():
            idx = library_idxs[lib_id]
            with open(in_filename, 'rt') as in_file:
                for line in in_file:
                    line = str(idx) + line[line.index('\t'):]
                    out_file.write(line)


def merge_sorted_files_by_line(in_filenames, out_filename, temp_space, sort_fields):
    try:
        os.makedirs(temp_space)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    pypeliner.commandline.execute(*['sort', '-T', temp_space, '-m', '-n', '-k', sort_fields] + list(in_filenames.values()) + ['>', out_filename])


def generate_chromosome_args(chromosomes):
    args = list()
    for chromosome_pair in itertools.combinations_with_replacement(chromosomes, 2):
        args.append('--inclchrompair ' + ','.join(chromosome_pair))
    args.append('--exclchrompairs ' + ','.join(chromosomes))
    return dict(enumerate(args))


def read_clusters_breakpoints(clusters_filename, breakpoints_filename):
    with open(clusters_filename, 'rt') as clusters_file, open(breakpoints_filename, 'rt') as breakpoints_file:
        clusters_reader = csv.reader(clusters_file, delimiter='\t')
        breakpoints_reader = csv.reader(breakpoints_file, delimiter='\t')
        cluster_iter = itertools.groupby(clusters_reader, lambda row: row[0])
        breakend_iter = itertools.groupby(breakpoints_reader, lambda row: row[0])
        for (cluster_id_1, cluster_rows), (cluster_id_2, breakend_rows) in zip(cluster_iter, breakend_iter):
            if cluster_id_1 != cluster_id_2:
                raise ValueError('Consistency issue between clusters and breakpoints for ' + clusters_filename + ' and ' + breakpoints_filename)
            yield cluster_id_1, cluster_rows, breakend_rows


def merge_clusters(in_clusters_filenames, in_breakpoints_filenames,
                   out_clusters_filename, out_breakpoints_filename, debug_filename):
    new_cluster_id = 0
    with open(out_clusters_filename, 'wt') as out_clusters_file, \
         open(out_breakpoints_filename, 'wt') as out_breakpoints_file, \
         open(debug_filename, 'wt') as debug_file:
        for idx, in_clusters_filename in in_clusters_filenames.items():
            in_breakpoints_filename = in_breakpoints_filenames[idx]
            for cluster_id, cluster_rows, breakend_rows in read_clusters_breakpoints(in_clusters_filename, in_breakpoints_filename):
                for row in cluster_rows:
                    row[0] = str(new_cluster_id)
                    out_clusters_file.write('\t'.join(row) + '\n')
                for row in breakend_rows:
                    row[0] = str(new_cluster_id)
                    out_breakpoints_file.write('\t'.join(row) + '\n')
                debug_file.write('{0}\t{1}\t{2}\n'.format(new_cluster_id, idx, cluster_id))
                new_cluster_id += 1


def tabulate_reads(clusters_filename, likelihoods_filename, library_ids, reads1_filenames, reads2_filenames, reads_table_filename):
    fields = ['cluster_id', 'cluster_end', 'lib_id', 'read_id', 'read_end', 'align_id']
    clusters = pd.read_csv(clusters_filename, sep='\t', names=fields, usecols=['cluster_id', 'lib_id', 'read_id'])
    clusters = clusters.drop_duplicates().set_index(['lib_id', 'read_id']).sort_index()['cluster_id']

    # The likelihoods file contains a list of read alignments that have passed
    # filtering, use this to generate a list of library / read id pairs that
    # can be used to annotate each read cluster assignment as filtered or not
    likelihoods = pd.read_csv(likelihoods_filename, sep='\t',
                              names=destruct.predict_breaks.likelihoods_fields,
                              converters=converters)
    passed_reads = set(zip(likelihoods.library_id, likelihoods.read_id))

    with open(reads_table_filename, 'wt') as reads_table_file:
        for lib_name in set(reads1_filenames.keys()).union(set(reads2_filenames.keys())):
            lib_id = library_ids[lib_name]
            for reads_filename in [reads1_filenames[lib_name], reads2_filenames[lib_name]]:
                with gzip.open(reads_filename, 'rt') as reads_file:
                    for name, seq, comment, qual in itertools.zip_longest(*[(a.rstrip() for a in reads_file)]*4):
                        assert name[0] == '@'
                        assert name[-1] == '1' or name[-1] == '2'
                        assert name[-2] == '/'
                        fragment_id = int(name[1:-2])
                        read_end = name[-1]
                        if (lib_id, fragment_id) in clusters.index:
                            cluster_id = clusters.loc[(lib_id, fragment_id)]
                            assert np.issubdtype(type(cluster_id), np.integer)
                            filtered = 'False' if (lib_id, fragment_id) in passed_reads else 'True'
                            reads_table_file.write('\t'.join([str(cluster_id), str(lib_id), str(fragment_id), read_end, seq, qual, comment, filtered]) + '\n')


class DGVDatabase(object):
    def __init__(self, dgv_filename):
        self.variations = list()
        chrvars = collections.defaultdict(list)
        with open(dgv_filename, 'rt') as dgv_file:
            dgv_reader = csv.reader(dgv_file, delimiter='\t')
            dgv_header = next(dgv_reader)
            for row in dgv_reader:
                id = row[0]
                chr = row[1]
                start = int(row[2])
                end = int(row[3])
                idx = len(self.variations)
                chrvars[chr].append((idx, start, end))
                self.variations.append((id, start, end))
        self.intervals = dict()
        for chr, vars in chrvars.items():
            self.intervals[chr] = pygenes.IntervalTree(vars)

    def query(self, chromosome, start, end):
        if chromosome not in self.intervals:
            return
        idxs = self.intervals[chromosome].find_overlapping(start, end)
        for idx in [idxs[a] for a in range(0, len(idxs))]:
            startdiff = abs(start - self.variations[idx][1])
            enddiff = abs(end - self.variations[idx][2])
            if startdiff < 500 and enddiff < 500:
                yield self.variations[idx][0]


def merge_tars(output_filename, *input_filename_sets):
    with tarfile.open(output_filename, 'wt') as output_tar:
        for input_filenames in input_filename_sets:
            for input_filename in input_filenames.values():
                with tarfile.open(input_filename, 'rt') as in_tar:
                    for tarinfo in in_tar:
                        output_tar.addfile(tarinfo, in_tar.extractfile(tarinfo))


converters = {'chromosome': str,
              'chromosome_1': str,
              'chromosome_2': str,
              'inserted': str}


def create_sequence(row, reference_sequences):
    breakend_sequences = ['', '']
    expected_strands = ('+', '-')
    inserted = ''
    if inserted != '.':
        inserted = row['inserted']
    for side in (0, 1):
        chromosome = row['chromosome_{0}'.format(side+1)]
        strand = row['strand_{0}'.format(side+1)]
        position = row['position_{0}'.format(side+1)]
        length = row['template_length_{0}'.format(side+1)]
        if strand == '+':
            start = position - length + 1
            end = position
        else:
            start = position
            end = position + length - 1
        breakend_sequences[side] = reference_sequences[chromosome][start-1:end]
        if strand != expected_strands[side]:
            breakend_sequences[side] = destruct.utils.misc.reverse_complement(breakend_sequences[side])
    return breakend_sequences[0] + '[' + inserted + ']' + breakend_sequences[1]


def annotate_genes(row, gene_models):

    for side in (0, 1):

        chromosome = row['chromosome_{0}'.format(side+1)]
        position = row['position_{0}'.format(side+1)]

        nearest_gene_ids = gene_models.find_nearest_genes(chromosome, position)
        gene_id = 'NA'
        gene_name = 'NA'
        gene_location = 'NA'
        if len(nearest_gene_ids) > 0:
            gene_id = nearest_gene_ids[0]
            gene_name = gene_models.get_gene(gene_id).name
            gene_location = gene_models.calculate_gene_location(gene_id, position)

        row['gene_id_{0}'.format(side+1)] = gene_id
        row['gene_name_{0}'.format(side+1)] = gene_name
        row['gene_location_{0}'.format(side+1)] = gene_location

    return row


def query_dgv(row, dgv):

    if row['chromosome_1'] != row['chromosome_2']:
        return 'NA'

    chromosome = row['chromosome_1']
    start, end = sorted((row['position_1'], row['position_2']))

    variants = list(dgv.query(chromosome, start, end))

    if len(variants) == 0:
        return 'NA'

    return ', '.join(variants)


def tabulate_results(breakpoints_filename, likelihoods_filename, library_ids,
                     genome_fasta, gtf_filename, dgv_filename,
                     breakpoint_table, breakpoint_library_table):

    lib_names = pd.DataFrame(library_ids.items(), columns=['library', 'library_id'])

    breakpoints = pd.read_csv(breakpoints_filename, sep='\t',
                              names=destruct.predict_breaks.breakpoint_fields,
                              converters=converters)
    breakpoints = breakpoints.drop(['breakpoint_id'], axis=1)
    breakpoints = breakpoints.rename(columns={'count':'num_split'})
    breakpoints.loc[breakpoints['inserted'] == '.', 'inserted'] = ''

    likelihoods = pd.read_csv(likelihoods_filename, sep='\t',
                              names=destruct.predict_breaks.likelihoods_fields,
                              converters=converters)
    likelihoods = likelihoods.drop(['breakpoint_id'], axis=1)

    breakpoint_reads = (
        likelihoods.groupby(['cluster_id', 'library_id'])
        .size()
        .reset_index()
    )
    breakpoint_reads.columns = ['cluster_id', 'library_id', 'num_reads']

    breakpoint_unique_reads = (
        likelihoods.drop_duplicates(['cluster_id', 'library_id', 'template_length_1', 'template_length_2'])
        .groupby(['cluster_id', 'library_id'])
        .size()
        .reset_index()
    )
    breakpoint_unique_reads.columns = ['cluster_id', 'library_id', 'num_unique_reads']

    breakpoint_library = (
        breakpoint_reads.merge(breakpoint_unique_reads)
        .merge(lib_names)
        .drop(['library_id'], axis=1)
    )

    agg_f = {
        'log_likelihood':np.average,
        'log_cdf':np.average,
        'template_length_1':max,
        'template_length_2':max,
    }

    breakpoint_stats = (
        likelihoods.groupby('cluster_id')
        .agg(agg_f)
        .reset_index()
    )

    breakpoint_stats['template_length_min'] = breakpoint_stats[['template_length_1', 'template_length_2']].min(axis=1)

    breakpoint_counts = (
        likelihoods.groupby('cluster_id')
        .size()
        .reset_index()
    )
    breakpoint_counts.columns = ['cluster_id', 'num_reads']

    breakpoint_unique_counts = (
        likelihoods.drop_duplicates(['cluster_id', 'library_id', 'template_length_1', 'template_length_2'])
        .groupby('cluster_id')
        .size()
        .reset_index()
    )
    breakpoint_unique_counts.columns = ['cluster_id', 'num_unique_reads']

    breakpoints = breakpoints.merge(breakpoint_stats, on='cluster_id', how='inner')
    breakpoints = breakpoints.merge(breakpoint_counts, on='cluster_id', how='inner')
    breakpoints = breakpoints.merge(breakpoint_unique_counts, on='cluster_id', how='inner')

    # Calculate breakpoint type
    def breakpoint_type(row):
        if row['chromosome_1'] != row['chromosome_2']:
            return 'translocation'
        if row['strand_1'] == row['strand_2']:
            return 'inversion'
        positions = sorted([(row['position_{0}'.format(side)], row['strand_{0}'.format(side)]) for side in (1, 2)])
        if positions[0][1] == '+':
            return 'deletion'
        else:
            return 'duplication'

    breakpoints['type'] = breakpoints.apply(breakpoint_type, axis=1)

    # Calculate number inserted at the breakpoint
    def calculate_num_inserted(row):
        if row['inserted'] == '.':
            return 0
        else:
            return len(row['inserted'])

    breakpoints['num_inserted'] = breakpoints.apply(calculate_num_inserted, axis=1)

    # Annotate sequence
    reference_sequences = dict()
    for id, seq in destruct.utils.seq.read_sequences(open(genome_fasta, 'rt')):
        reference_sequences[id] = seq

    breakpoints['sequence'] = breakpoints.apply(lambda row: create_sequence(row, reference_sequences), axis=1)

    # Annotate gene information
    gene_models = pygenes.GeneModels()
    gene_models.load_ensembl_gtf(gtf_filename)

    breakpoints = breakpoints.apply(lambda row: annotate_genes(row, gene_models), axis=1)

    # Annotate database of genomic variants
    dgv = DGVDatabase(dgv_filename)

    breakpoints['dgv_ids'] = breakpoints.apply(lambda row: query_dgv(row, dgv), axis=1)

    breakpoints = breakpoints.rename(columns={'cluster_id':'prediction_id'})

    breakpoints.to_csv(breakpoint_table, sep='\t', na_rep='NA', header=True, index=False)

    breakpoint_library = breakpoint_library.rename(columns={'cluster_id':'prediction_id'})

    breakpoint_library.to_csv(breakpoint_library_table, sep='\t', na_rep='NA', header=True, index=False)



